/*
 * Copyright (C) 2013 Texas Instruments Incorporated - http://www.ti.com/
 *
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <malloc.h>
#include <errno.h>

#include "mpm_transport_hyplnk.h"
#include "mpm_transport_hyplnk_interface.h"

#define MAX_DEVICE_NAME_LEN 32

/* Only one set of config can be active at one time. Keep track of it globally */
hyplnk_peripheral hyplnkActiveIF[NUM_INTERFACES];

int edma3InstInit[NUM_EDMA3_INSTANCES];

static int mpm_transport_get_global_addr (mpm_transport_cfg_t *sp, uint32_t laddr, uint32_t size, uint32_t *gaddr, int *index)
{
	int i;
	
	if ((CSL_MSMC_CFG_REGS <= laddr) && (laddr + size <= CSL_MSMC_CFG_REGS + 0x100000) ) {
		*gaddr = laddr;
		return 0;
	}

	for (i = 0; i < sp->num_mmap; i++) {
		if (sp->mmap[i].local_addr <= laddr &&
		    sp->mmap[i].local_addr + sp->mmap[i].length >= laddr + size) {
			*gaddr = sp->mmap[i].global_addr + laddr - sp->mmap[i].local_addr;
			*index = i;
			return 0;
		}
	}

	for (i = 0; i < sp->num_mmap; i++) {	
		if (sp->mmap[i].global_addr <= laddr &&
		    sp->mmap[i].global_addr + sp->mmap[i].length >= laddr + size) {
			*gaddr = laddr;
			*index = i;
			return 0;
		}
	}

	printf("can't find global address from local address 0x%x size %d in cfg for %s\n",
	       laddr, size, sp->name);
	return -1;
}

int mpm_transport_hyplnk_unregister_slave(mpm_transport_cfg_t *sp)
{
	mpm_transport_hyplnk_t *td = (mpm_transport_hyplnk_t *) sp->td;
	if (hyplnkActiveIF[td->hyplnkIF].slavesActive == 1)
	{
		strcpy(hyplnkActiveIF[td->hyplnkIF].name, "");
		hyplnkActiveIF[td->hyplnkIF].slavesActive = 0;
		hyplnkActiveIF[td->hyplnkIF].hyplnkSegments = 0;
		return 1;
	}
	else
	{
		hyplnkActiveIF[td->hyplnkIF].slavesActive -= 1;
		return 0;
	}
	return 0;
}

int mpm_transport_hyplnk_register_slave(mpm_transport_cfg_t *sp)
{
	mpm_transport_hyplnk_t *td = (mpm_transport_hyplnk_t *) sp->td;
	if (strlen(hyplnkActiveIF[td->hyplnkIF].name) < 2)
	{
		// First time init for this slave. Populate hyplnkActiveIF.
		hyplnkActiveIF[td->hyplnkIF].slavesActive = 1;
		hyplnkActiveIF[td->hyplnkIF].hyplnkSegments = 0;
		if (strcmp(td->hyplnkSettings,"")==0)
			strcpy(hyplnkActiveIF[td->hyplnkIF].name, sp->name);
		else
			strcpy(hyplnkActiveIF[td->hyplnkIF].name, td->hyplnkSettings);
		debug_printf("Slave %s registered setting name %s\n",sp->name, hyplnkActiveIF[td->hyplnkIF].name);
		return 1;
	}
	else
	{
		if (strcmp(hyplnkActiveIF[td->hyplnkIF].name, sp->name) == 0)
		{
			// Another slave already used you as the master
			hyplnkActiveIF[td->hyplnkIF].slavesActive += 1;
			return 0;
		}
		else
		{
			if (strcmp(hyplnkActiveIF[td->hyplnkIF].name, td->hyplnkSettings) == 0)
			{
				// hyplnkmaster already opened
				hyplnkActiveIF[td->hyplnkIF].slavesActive += 1;
				return 0;
			}
			else
			{
				// This hyplnk port is already opened with another set of settings
				printf("Error! Close slave: %s!\n",hyplnkActiveIF[td->hyplnkIF].name);
				return -1;
			}
		}
	}
	return 0;
}

int mpm_transport_hyplnk_open(mpm_transport_cfg_t *sp, mpm_transport_open_t *ocfg)
{
	int index = 0, reg, ret = 0;
	mpm_transport_hyplnk_t *td = (mpm_transport_hyplnk_t *) sp->td;

	reg = mpm_transport_hyplnk_register_slave(sp);
	if (reg < 0) return -1;

	index = td->hyplnkIF;
	if ((hyplnkActiveIF[index].slavesActive == 1) && (hyplnkActiveIF[((~index)&0x1)].slavesActive == 0))
	{
		hyplnkActiveIF[0].hyp_fd = open("/dev/hyperlink0", ocfg->open_mode);
		hyplnkActiveIF[1].hyp_fd = open("/dev/hyperlink1", ocfg->open_mode);
	}
	td->fd[0] = hyplnkActiveIF[0].hyp_fd;
	td->fd[1] = hyplnkActiveIF[1].hyp_fd;

		debug_printf("\n");
		debug_printf("========Hyperlnk Setup Info==========\n");
	if (td->hyplnkIF == hyplnk_PORT_0)
		debug_printf("= Port:        hyplnk0              =\n");
	else if (td->hyplnkIF == hyplnk_PORT_1)
		debug_printf("= Port:        hyplnk1              =\n");
	if (td->hyplnkDir == hyplnk_LOOPBACK)
		debug_printf("= Direction:   Loopback             =\n");
	else if (td->hyplnkDir == hyplnk_REMOTE)
		debug_printf("= Direction:   Remote               =\n");
	if (td->hyplnkRefClk == hyplnk_REFCLK_156p25)
		debug_printf("= Ref Clock:   156.25 MHz           =\n");
	else if (td->hyplnkRefClk == hyplnk_REFCLK_312p5)
		debug_printf("= Ref Clock:   312.5 MHz            =\n");
	if (td->hyplnkSerRate == hyplnk_SERRATE_6p25)
		debug_printf("= Serial Rate: 6.25 Gbps            =\n");
	if (td->hyplnkLaneRate == hyplnk_LANE_FULL)
		debug_printf("= Lane Rate:   FULL                 =\n");
	else if (td->hyplnkLaneRate == hyplnk_LANE_HALF)
		debug_printf("= Lane Rate:   HALF                 =\n");
	else if (td->hyplnkLaneRate == hyplnk_LANE_QUARTER)
		debug_printf("= Lane Rate:   QUARTER              =\n");
		debug_printf("=====================================\n\n");
	
	if (reg)
	{
		if (mpm_transport_hyplnk_init_peripheral(td, index, ocfg->msec_timeout) != 0)
		{
			printf("mpm_transport_hyplnk_init_peripheral failed!\n");
			ret = -1;
		}

		if (strlen(sp->edma3_name) > 1)
		{
			if (edma3InstInit[sp->edma3_inst] == 0)
			{
				debug_printf("Initialzing edma3 for instance %d\n",sp->edma3_inst);
				edma3InstInit[sp->edma3_inst] = 1;
				mpm_transport_edma3_init_cfg(sp->edma3_inst);
			}

			sp->hEdma[sp->edma3_inst] = mpm_transport_edma3_setup(sp->edma3_inst, sp->shadow_reg);
			debug_printf("Got hEdma = 0x%08x\n",(uint32_t)sp->hEdma[sp->edma3_inst]);
			if (sp->hEdma[sp->edma3_inst] == 0)
			{
				printf("mpm_transport_edma3_setup failed!\n");
				ret = -1;
			}
		}
	}

	if (ret)
	{
		printf("Peripheral init failed for: %s\n", sp->name);
		mpm_transport_hyplnk_unregister_slave(sp);
		mpm_transport_hyplnk_reset(td, index);
		return -1;
	}
	return 0;

}

uint32_t mpm_transport_hyplnk_window_map(mpm_transport_cfg_t *sp, uint32_t addr, uint32_t length, mpm_transport_mmap_t *mcfg)
{
#if TIME_PROFILE_HYPLNK_WINDOW_MAP
	struct timespec tp_start, tp_end;
#endif
	int index;
	int user_index;
	uint32_t gaddr;
	uint32_t total_length_used = 0;
	uint32_t hlink_offset;
	uint8_t seg_idx = 0;
	mpm_transport_hyplnk_t *td = (mpm_transport_hyplnk_t *) sp->td;

#if TIME_PROFILE_HYPLNK_WINDOW_MAP
	clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
	/* Find a free mmap index to use */
	for (user_index = 0; user_index < MPM_MAX_USER_MMAPS; user_index++) {
		if (!td->mmap_user[user_index].addr) {
			break;
		}
		total_length_used += td->mmap_user[user_index].size;
	}

	if (user_index >= MPM_MAX_USER_MMAPS) {
		printf("exceeded maximum user mmap limits (%d)\n",
 		       MPM_MAX_USER_MMAPS);
		return -1;
	}

	if(mpm_transport_get_global_addr(sp, addr, length, &gaddr, &index)) {
		printf("can't find global address for local address 0x%x\n", addr);
		return -1;
	}

	gaddr = addr;
	for (index=0; index<32; index++) {
		if ((hyplnkActiveIF[td->hyplnkIF].hyplnkSegments & (1<<index)) == 0) {
			seg_idx = index;
			break;
		}
	}

#if TIME_PROFILE_HYPLNK_WINDOW_MAP
	clock_gettime(CLOCK_MONOTONIC, &tp_end);
	printf("[TIME PROFILE: mpm_transport_hyplnk_window_map] Time taken for setup: %lld\n",clock_diff (&tp_start, &tp_end));
#endif

#if TIME_PROFILE_HYPLNK_WINDOW_MAP
	clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
	td->mmap_user[user_index].addr = mpm_transport_hyplnk_get_window(td, gaddr, seg_idx, &hlink_offset);
	if (td->mmap_user[user_index].addr == NULL) {
		printf("can't mmap for the address 0x%x with length 0x%x (err: %s)\n",
 		       addr, length, strerror(errno));
		return -1;
	}
	else {
		hyplnkActiveIF[td->hyplnkIF].hyplnkSegments |= (1<<index);
	}
#if TIME_PROFILE_HYPLNK_WINDOW_MAP
	clock_gettime(CLOCK_MONOTONIC, &tp_end);
	printf("[TIME PROFILE: mpm_transport_hyplnk_window_map] Time taken for get_window(): %lld\n",clock_diff (&tp_start, &tp_end));
#endif
	td->mmap_user[user_index].hyplnk_idx = seg_idx;
	td->mmap_user[user_index].size = length;
	td->mmap_user[user_index].addr_usr = td->mmap_user[user_index].addr + hlink_offset;
	td->mmap_user[user_index].phy_addr = hyplnkActiveIF[td->hyplnkIF].windowBase + hlink_offset;

	return user_index;
}

void *mpm_transport_hyplnk_mmap(mpm_transport_cfg_t *sp, uint32_t addr, uint32_t length, mpm_transport_mmap_t *mcfg)
{
	uint32_t index;
	mpm_transport_hyplnk_t *td = (mpm_transport_hyplnk_t *) sp->td;
	index = mpm_transport_hyplnk_window_map(sp, addr, length, mcfg);
	if (index < 0)
	{
		printf("hyperlink window mapping error!\n");
		return 0;
	}
	debug_printf("[debug] index is %d\n",index);
	debug_printf("[debug] addr is 0x%08x\n", (uint32_t) td->mmap_user[index].addr);
	debug_printf("[debug] addr_usr is 0x%08x\n",(uint32_t) td->mmap_user[index].addr_usr);
	debug_printf("[debug] phy_addr is 0x%08x\n",td->mmap_user[index].phy_addr);
	return (td->mmap_user[index].addr_usr);
}

uint32_t mpm_transport_hyplnk_phys_map(mpm_transport_cfg_t *sp, uint32_t addr, uint32_t length, mpm_transport_mmap_t *mcfg)
{
	uint32_t index;
	mpm_transport_hyplnk_t *td = (mpm_transport_hyplnk_t *) sp->td;
	index = mpm_transport_hyplnk_window_map(sp, addr, length, mcfg);
	if (index < 0)
	{
		printf("hyperlink window mapping error!\n");
		return 0;
	}
	debug_printf("[debug] index is %d\n",index);
	debug_printf("[debug] addr is 0x%08x\n", (uint32_t) td->mmap_user[index].addr);
	debug_printf("[debug] addr_usr is 0x%08x\n",(uint32_t) td->mmap_user[index].addr_usr);
	debug_printf("[debug] phy_addr is 0x%08x\n",td->mmap_user[index].phy_addr);
	return (td->mmap_user[index].phy_addr);
}

int mpm_transport_hyplnk_munmap(mpm_transport_cfg_t *sp, void *va, uint32_t length)
{
	int user_index;
	mpm_transport_hyplnk_t *td = (mpm_transport_hyplnk_t *) sp->td;
	
	for (user_index = 0; user_index < MPM_MAX_USER_MMAPS; user_index++) {
		if (td->mmap_user[user_index].addr == va) {
			break;
		}
	}

	if (user_index >= MPM_MAX_USER_MMAPS) {
		printf("can't find mapping to unmap\n");
		return -1;
	}
	hyplnkActiveIF[td->hyplnkIF].hyplnkSegments &= ~(1<<td->mmap_user[user_index].hyplnk_idx);
	td->mmap_user[user_index].addr = 0;
	td->mmap_user[user_index].size = 0;
	td->mmap_user[user_index].addr_usr = 0;
	td->mmap_user[user_index].phy_addr = 0;
	td->mmap_user[user_index].logical_addr = 0;
	td->mmap_user[user_index].hyplnk_idx = 0;
	return 0;
}

int mpm_transport_hyplnk_phys_unmap(mpm_transport_cfg_t *sp, void *va, uint32_t length)
{
	int user_index, logical_idx, logical_addr;
	mpm_transport_hyplnk_t *td = (mpm_transport_hyplnk_t *) sp->td;
	
	for (user_index = 0; user_index < MPM_MAX_USER_MMAPS; user_index++) {
		if (td->mmap_user[user_index].phy_addr == (uint32_t) va) {
			break;
		}
	}

	if (user_index >= MPM_MAX_USER_MMAPS) {
		printf("can't find mapping to unmap\n");
		return -1;
	}
	logical_addr = td->mmap_user[user_index].logical_addr;
	if (logical_addr)
	{
		for (logical_idx = 0; logical_idx < MPM_MAX_USER_MMAPS; logical_idx++) {
			if (td->mmap_user[logical_idx].logical_addr == logical_addr) {
				break;
			}
		}
		mpm_transport_keystone_unmap(sp->rhMpax,logical_addr,sp->mpax_size);
		hyplnkActiveIF[td->hyplnkIF].hyplnkSegments &= ~(1<<td->mmap_user[logical_idx].hyplnk_idx);
		td->mmap_user[logical_idx].addr = 0;
		td->mmap_user[logical_idx].size = 0;
		td->mmap_user[logical_idx].addr_usr = 0;
		td->mmap_user[logical_idx].phy_addr = 0;
		td->mmap_user[logical_idx].logical_addr = 0;
		td->mmap_user[logical_idx].hyplnk_idx = 0;
	}

	hyplnkActiveIF[td->hyplnkIF].hyplnkSegments &= ~(1<<td->mmap_user[user_index].hyplnk_idx);
	td->mmap_user[user_index].addr = 0;
	td->mmap_user[user_index].size = 0;
	td->mmap_user[user_index].addr_usr = 0;
	td->mmap_user[user_index].phy_addr = 0;
	td->mmap_user[user_index].logical_addr = 0;
	td->mmap_user[user_index].hyplnk_idx = 0;
	return 0;
}

void *mpm_transport_hyplnk_rw_map (mpm_transport_cfg_t *sp, uint32_t addr, uint32_t length, int *index)
{
	mpm_transport_hyplnk_t *td = (mpm_transport_hyplnk_t *) sp->td;
	mpm_transport_mmap_t mcfg = {
		.mmap_prot	= (PROT_READ|PROT_WRITE),
		.mmap_flags	= MAP_SHARED,
	};

	*index = mpm_transport_hyplnk_window_map(sp, addr, length, &mcfg);
	if (*index < 0)
	{
		printf("hyperlink window mapping error!\n");
		return 0;
	}
	return (td->mmap_user[*index].addr_usr);
}

int mpm_transport_hyplnk_read(mpm_transport_cfg_t *sp, uint32_t addr, uint32_t length, char *buf, mpm_transport_read_t *rcfg)
{
	int index = 0;
	void *mmap_addr;
	mpm_transport_hyplnk_t *td = (mpm_transport_hyplnk_t *) sp->td;

	mmap_addr = mpm_transport_hyplnk_rw_map (sp, addr, length, &index);

	if (!mmap_addr) {
		printf("can't map for address 0x%x\n", addr);
		return -1;
	}
	memcpy(buf, mmap_addr, length);
	mpm_transport_hyplnk_munmap(sp, td->mmap_user[index].addr, length);
	return 0;
}

int mpm_transport_hyplnk_write(mpm_transport_cfg_t *sp, uint32_t addr, uint32_t length, char *buf, mpm_transport_write_t *cfg)
{
	int index = 0;
	void *mmap_addr;
	mpm_transport_hyplnk_t *td = (mpm_transport_hyplnk_t *) sp->td;

	mmap_addr = mpm_transport_hyplnk_rw_map (sp, addr, length, &index);

	if (!mmap_addr) {
		printf("can't map for address 0x%x\n", addr);
		return -1;
	}
	memcpy(mmap_addr, buf, length);
	mpm_transport_hyplnk_munmap(sp, td->mmap_user[index].addr, length);
	return 0;
}

mpm_transport_trans_h mpm_transport_hyplnk_get_initiate(mpm_transport_cfg_t *sp, uint32_t from_addr, uint32_t to_addr, uint32_t length, bool is_blocking, mpm_transport_read_t *cfg)
{
#if TIME_PROFILE_HYPLNK_GET_INITIATE
	struct timespec tp_start, tp_end;
#endif
	mpm_transport_trans_t *tp = NULL;
	int i;
	void *mmap_addr;
	mpm_transport_mmap_t mcfg = {
		.mmap_prot	= (PROT_READ|PROT_WRITE),
		.mmap_flags	= MAP_SHARED,
	};
	for (i = 0; i < MPM_MAX_DMA_TRANS; i++)
	{
		if (sp->ta[i].is_active == 0)
			break;
	}
	if (i == MPM_MAX_DMA_TRANS)
	{
		printf("MAX DMA Transfers! Call mpm_transport_transfer_check to wait/ccomplete pending transfers\n");
		tp->key = -1;
		return tp;
	}
	tp = (mpm_transport_trans_t *) &(sp->ta[i]);
	tp->key =  (i+3) * (i+3) * 2;
	tp->is_active = 1;

#if TIME_PROFILE_HYPLNK_GET_INITIATE
	clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
	mmap_addr = (void *) mpm_transport_hyplnk_phys_map (sp, from_addr, length, &mcfg);
	if (!mmap_addr) {
		printf("can't map for address 0x%x\n", to_addr);
		return (void *) -1;
	}
#if TIME_PROFILE_HYPLNK_GET_INITIATE
	clock_gettime(CLOCK_MONOTONIC, &tp_end);
	printf("[TIME PROFILE: mpm_transport_hyplnk_get_initiate] Time taken for phys_map: %lld\n",clock_diff (&tp_start, &tp_end));
#endif

	tp->trans_addr = (uint32_t) mmap_addr;

#if TIME_PROFILE_HYPLNK_GET_INITIATE
	clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
	debug_printf("mpm_transport_get_initiate: getting address 0x%08x, to 0x%08x\n", from_addr, to_addr);
	mpm_transport_edma3_transfer(sp->hEdma[sp->edma3_inst], (uint32_t) mmap_addr, to_addr, length, is_blocking, tp);
	debug_printf("hEdma = 0x%08x, Channel id: 0x%08x\n", (uint32_t) sp->hEdma[sp->edma3_inst], tp->chan_id);
#if TIME_PROFILE_HYPLNK_GET_INITIATE
	clock_gettime(CLOCK_MONOTONIC, &tp_end);
	printf("[TIME PROFILE: mpm_transport_hyplnk_put_initiate] Time taken for edma3_transfer(): %lld\n",clock_diff (&tp_start, &tp_end));
#endif
	if (is_blocking == true) mpm_transport_hyplnk_phys_unmap(sp, mmap_addr, 0);
	return tp;
}

mpm_transport_trans_h mpm_transport_hyplnk_put_initiate(mpm_transport_cfg_t *sp, uint32_t to_addr, uint32_t from_addr, uint32_t length, bool is_blocking, mpm_transport_write_t *cfg)
{
#if TIME_PROFILE_HYPLNK_PUT_INITIATE
	struct timespec tp_start, tp_end;
#endif
	mpm_transport_trans_t *tp = NULL;
	int i;
	void *mmap_addr;
	mpm_transport_mmap_t mcfg = {
		.mmap_prot	= (PROT_READ|PROT_WRITE),
		.mmap_flags	= MAP_SHARED,
	};
	for (i = 0; i < MPM_MAX_DMA_TRANS; i++)
	{
		if (sp->ta[i].is_active == 0)
			break;
	}
	if (i == MPM_MAX_DMA_TRANS)
	{
		printf("MAX DMA Transfers! Call mpm_transport_transfer_check to wait/ccomplete pending transfers\n");
		tp->key = -1;
		return tp;
	}
	tp = (mpm_transport_trans_t *) &(sp->ta[i]);
	tp->key = (i+1) * (i+1) * 2;
	tp->is_active = 1;

#if TIME_PROFILE_HYPLNK_PUT_INITIATE
	clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
	mmap_addr = (void *) mpm_transport_hyplnk_phys_map (sp, to_addr, length, &mcfg);
	if (!mmap_addr) {
		printf("can't map for address 0x%x\n", to_addr);
		return (void *) -1;
	}
	tp->trans_addr = (uint32_t) mmap_addr;
#if TIME_PROFILE_HYPLNK_PUT_INITIATE
	clock_gettime(CLOCK_MONOTONIC, &tp_end);
	printf("[TIME PROFILE: mpm_transport_hyplnk_put_initiate] Time taken for phys_map: %lld\n",clock_diff (&tp_start, &tp_end));
#endif

#if TIME_PROFILE_HYPLNK_PUT_INITIATE
	clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
	debug_printf("mpm_transport_put_initiate: writing to address 0x%08x, from 0x%08x\n", to_addr, from_addr);
	mpm_transport_edma3_transfer(sp->hEdma[sp->edma3_inst], from_addr, (uint32_t) mmap_addr, length, is_blocking, tp);
	debug_printf("hEdma = 0x%08x, Channel id: 0x%08x\n", (uint32_t) sp->hEdma[sp->edma3_inst], tp->chan_id);
#if TIME_PROFILE_HYPLNK_PUT_INITIATE
	clock_gettime(CLOCK_MONOTONIC, &tp_end);
	printf("[TIME PROFILE: mpm_transport_hyplnk_put_initiate] Time taken for edma3_transfer(): %lld\n",clock_diff (&tp_start, &tp_end));
#endif

	if (is_blocking == true) mpm_transport_hyplnk_phys_unmap(sp, mmap_addr, 0);
	return tp;
}

mpm_transport_trans_h mpm_transport_hyplnk_get_initiate_linked(mpm_transport_cfg_t *sp, uint32_t num_links, uint32_t from_addr[], uint32_t to_addr[], uint32_t length[], bool is_blocking, mpm_transport_write_t *cfg)
{
#if TIME_PROFILE_HYPLNK_GET_INITIATE_LINKED
	struct timespec tp_start, tp_end;
#endif
	mpm_transport_trans_t *tp = NULL;
	mpm_transport_trans_t *trav = NULL;
	int i;
	uint32_t mmap_addr[num_links];
	mpm_transport_mmap_t mcfg = {
		.mmap_prot	= (PROT_READ|PROT_WRITE),
		.mmap_flags	= MAP_SHARED,
	};
	for (i = 0; i < MPM_MAX_DMA_TRANS; i++)
	{
		if (sp->ta[i].is_active == 0)
			break;
	}
	if (i == MPM_MAX_DMA_TRANS)
	{
		printf("MAX DMA Transfers! Call mpm_transport_transfer_check to wait/ccomplete pending transfers\n");
		tp->key = -1;
		return tp;
	}
	tp = (mpm_transport_trans_t *) &(sp->ta[i]);
	trav = tp;
	tp->key =  (i+3) * (i+3) * 2;
	tp->is_active = 1;

	for (i = 0; i<num_links; i++)
	{
#if TIME_PROFILE_HYPLNK_GET_INITIATE_LINKED
	clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
		mmap_addr[i] = (uint32_t) mpm_transport_hyplnk_phys_map (sp, from_addr[i], length[i], &mcfg);
		if (!mmap_addr[i]) {
			printf("can't map for address 0x%x\n", from_addr[i]);
			return (void *) -1;
		}
#if TIME_PROFILE_HYPLNK_GET_INITIATE_LINKED
	clock_gettime(CLOCK_MONOTONIC, &tp_end);
	printf("[TIME PROFILE: mpm_transport_hyplnk_get_initiate_linked] Time taken for phys_map: %lld, iteration: %d\n",clock_diff (&tp_start, &tp_end),i);
#endif
	}

#if TIME_PROFILE_HYPLNK_GET_INITIATE_LINKED
	clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
	mpm_transport_edma3_transfer_linked(sp->hEdma[sp->edma3_inst], num_links, mmap_addr, to_addr, length, is_blocking, tp);
#if TIME_PROFILE_HYPLNK_GET_INITIATE_LINKED
	clock_gettime(CLOCK_MONOTONIC, &tp_end);
	printf("[TIME PROFILE: mpm_transport_hyplnk_get_initiate_linked] Time taken for edma3_transfer_linked(): %lld\n",clock_diff (&tp_start, &tp_end));
#endif

	// Record the addresses used so we can unmap later
	i = 0;
	do {
		if (i<num_links) trav->trans_addr = (uint32_t) mmap_addr[i];
		else trav->trans_addr = 0;
		trav->local_logical_addr = 0;
		trav = (mpm_transport_trans_t *) trav->chain;
		i++;
	} while (trav != NULL);

	if (is_blocking == true) while(mpm_transport_hyplnk_transfer_check(sp, tp) != 1);
	return tp;	
}

mpm_transport_trans_h mpm_transport_hyplnk_put_initiate_linked(mpm_transport_cfg_t *sp, uint32_t num_links, uint32_t to_addr[], uint32_t from_addr[], uint32_t length[], bool is_blocking, mpm_transport_write_t *cfg)
{
#if TIME_PROFILE_HYPLNK_PUT_INITIATE_LINKED
	struct timespec tp_start, tp_end;
#endif
	mpm_transport_trans_t *tp = NULL;
	mpm_transport_trans_t *trav = NULL;
	int i;
	uint32_t mmap_addr[num_links];
	mpm_transport_mmap_t mcfg = {
		.mmap_prot	= (PROT_READ|PROT_WRITE),
		.mmap_flags	= MAP_SHARED,
	};
	for (i = 0; i < MPM_MAX_DMA_TRANS; i++)
	{
		if (sp->ta[i].is_active == 0)
			break;
	}
	if (i == MPM_MAX_DMA_TRANS)
	{
		printf("MAX DMA Transfers! Call mpm_transport_transfer_check to wait/ccomplete pending transfers\n");
		tp->key = -1;
		return tp;
	}
	tp = (mpm_transport_trans_t *) &(sp->ta[i]);
	trav = tp;
	tp->key = (i+1) * (i+1) * 2;
	tp->is_active = 1;

	for (i = 0; i<num_links; i++)
	{
#if TIME_PROFILE_HYPLNK_PUT_INITIATE_LINKED
	clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
		mmap_addr[i] = (uint32_t) mpm_transport_hyplnk_phys_map (sp, to_addr[i], length[i], &mcfg);
		if (!mmap_addr[i]) {
			printf("can't map for address 0x%x\n", to_addr[i]);
			return (void *) -1;
		}
#if TIME_PROFILE_HYPLNK_PUT_INITIATE_LINKED
	clock_gettime(CLOCK_MONOTONIC, &tp_end);
	printf("[TIME PROFILE: mpm_transport_hyplnk_put_initiate_linked] Time taken for phys_map: %lld, iteration: %d\n",clock_diff (&tp_start, &tp_end),i);
#endif
	}

#if TIME_PROFILE_HYPLNK_PUT_INITIATE_LINKED
	clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
	mpm_transport_edma3_transfer_linked(sp->hEdma[sp->edma3_inst], num_links, from_addr, mmap_addr, length, is_blocking, tp);
#if TIME_PROFILE_HYPLNK_PUT_INITIATE_LINKED
	clock_gettime(CLOCK_MONOTONIC, &tp_end);
	printf("[TIME PROFILE: mpm_transport_hyplnk_put_initiate_linked] Time taken for edma3_transfer_linked(): %lld\n",clock_diff (&tp_start, &tp_end));
#endif

	// Record the addresses used so we can unmap later
	i = 0;
	do {
		if (i<num_links) trav->trans_addr = (uint32_t) mmap_addr[i];
		else trav->trans_addr = 0;
		trav->local_logical_addr = 0;
		trav = (mpm_transport_trans_t *) trav->chain;
		i++;
	} while (trav != NULL);

	if (is_blocking == true) while(mpm_transport_hyplnk_transfer_check(sp, tp) != 1);
	return tp;
}

int mpm_transport_hyplnk_transfer_check(mpm_transport_cfg_t *sp, mpm_transport_trans_h th)
{
#if TIME_PROFILE_HYPLNK_TRANSFER_CHECK
	struct timespec tp_start, tp_end;
#endif
	mpm_transport_trans_t *tp = (mpm_transport_trans_t *) th;
	mpm_transport_trans_t *trav = tp;
	mpm_transport_trans_t *travBack = trav;
	int ret = 0;
	int i;

#if TIME_PROFILE_HYPLNK_TRANSFER_CHECK
	clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
	for (i = 0; i<MPM_MAX_DMA_TRANS; i++)
	{
		if (sp->ta[i].key == tp->key)
			break;
	}
	if (i == MPM_MAX_DMA_TRANS)
	{
		printf("Transaction handle with key 0x%08x not found!\n",tp->key);
		return ret;
	}
	debug_printf("Checking...: hEdma = 0x%08x, chan_id = 0x%08x, tcc_no = 0x%08x\n", (uint32_t) sp->hEdma[sp->edma3_inst], tp->chan_id, tp->tcc_no);
	ret = mpm_transport_edma3_checkAndClearTcc(sp->hEdma[sp->edma3_inst],trav->tcc_no);
#if TIME_PROFILE_HYPLNK_TRANSFER_CHECK
	clock_gettime(CLOCK_MONOTONIC, &tp_end);
	printf("[TIME PROFILE: mpm_transport_hyplnk_transfer_check] Time taken for edma3_checkAndClearTcc(): %lld\n",clock_diff (&tp_start, &tp_end));
#endif
	if (ret == 1)
	{

#if TIME_PROFILE_HYPLNK_TRANSFER_CHECK
	clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
		do {
			if (trav->trans_addr) mpm_transport_hyplnk_phys_unmap(sp, (void *) trav->trans_addr, 0);
			if (trav->local_logical_addr) {
				mpm_transport_keystone_unmap(sp->hMpax,trav->local_logical_addr,sp->mpax_size);
			}
			mpm_transport_edma3_free_channel(sp->hEdma[sp->edma3_inst], trav->chan_id);
			trav = (mpm_transport_trans_t *) trav->chain;
		} while (trav != NULL);

		// Destroying linked transfer chain
		trav = tp->chain;
		while (trav != NULL) {
			travBack = trav;
			trav = trav->chain;
			free(travBack);
		}
		
		debug_printf("Got index %d with key 0x%08x! clearing...\n",i,sp->ta[i].key);
		sp->ta[i].key = 0;
		sp->ta[i].is_active = 0;
		sp->ta[i].trans_addr = 0;
		sp->ta[i].local_logical_addr = 0;
		sp->ta[i].tcc_no = 0;
		sp->ta[i].chain = NULL;
#if TIME_PROFILE_HYPLNK_TRANSFER_CHECK
	clock_gettime(CLOCK_MONOTONIC, &tp_end);
	printf("[TIME PROFILE: mpm_transport_hyplnk_transfer_check] Time taken for clearing resources: %lld\n",clock_diff (&tp_start, &tp_end));
#endif
	}

	return ret;

}

void mpm_transport_hyplnk_close(mpm_transport_cfg_t *sp)
{
	int i;
	int index, reg;
	mpm_transport_hyplnk_t *td = (mpm_transport_hyplnk_t *) sp->td;

	if (!td) {
		return;
	}
	index = td->hyplnkIF;
	reg = mpm_transport_hyplnk_unregister_slave(sp);
	if (reg)
		mpm_transport_hyplnk_reset(td, index);

	if (strlen(sp->edma3_name) > 1) {
		mpm_transport_edma3_close(sp->hEdma[sp->edma3_inst], sp->edma3_inst);
	}

	if (sp->hMpax) {
		mpm_transport_keystone_free(sp->hMpax);
	}

	if (sp->rhMpax) {
		mpm_transport_keystone_free(sp->rhMpax);
	}
	
	for (i = 0; i < MPM_MAX_USER_MMAPS; i++) {
		if(td->mmap_user[i].addr) {
			memset(&(td->mmap_user[i]), 0, sizeof(mmap_data_t));
		}
	}
	for (i = 0; i < MPM_MAX_MMAPS; i++) {
		if(td->mmap_rw[i].addr) {
			memset(&(td->mmap_rw[i]), 0, sizeof(mmap_data_t));
		}
		pthread_mutex_destroy(&td->mutex_rw[i]);
	}
	if ((hyplnkActiveIF[0].slavesActive == 0) && (hyplnkActiveIF[1].slavesActive == 0))
	{
		close(hyplnkActiveIF[0].hyp_fd);
		close(hyplnkActiveIF[1].hyp_fd);
	}
	if (td) free(td);
}

uint32_t mpm_transport_hyplnk_phys_map64(mpm_transport_cfg_t *sp, uint64_t addr, uint32_t length, mpm_transport_mmap_t *mcfg)
{
	uint32_t index, logical_addr;
	mpm_transport_hyplnk_t *td = (mpm_transport_hyplnk_t *) sp->td;
	if (((addr>>32)&(0xFFFFFFFF)) == 0)
		index = mpm_transport_hyplnk_window_map(sp, addr, length, mcfg);
	else
	{
		index = mpm_transport_hyplnk_window_map(sp, CSL_MSMC_CFG_REGS, 0x100000, mcfg);
		if (!(sp->rhMpax)) {
			sp->rhMpax = mpm_transport_keystone_mmap_open(sp->mpax_base, sp->mpax_size, sp->mpax_index, sp->mpax_count, td->mmap_user[index].phy_addr, td->fd[td->hyplnkIF], 14);
		}
		logical_addr = mpm_transport_keystone_mmap(sp->rhMpax, addr, length);
		debug_printf("Got hyplnk-mapped msmc config logical_addr: 0x%08x\n",logical_addr);
		//mpm_transport_hyplnk_phys_unmap(sp, td->mmap_user[index].phy_addr, 0);
		td->mmap_user[index].logical_addr = logical_addr;
		index = mpm_transport_hyplnk_window_map(sp, logical_addr, length, mcfg);
		td->mmap_user[index].logical_addr = logical_addr;
	}
	if (index < 0)
	{
		printf("hyperlink window mapping error!\n");
		return 0;
	}
	debug_printf("[debug] index is %d\n",index);
	debug_printf("[debug] addr is 0x%08x\n", (uint32_t) td->mmap_user[index].addr);
	debug_printf("[debug] addr_usr is 0x%08x\n",(uint32_t) td->mmap_user[index].addr_usr);
	debug_printf("[debug] phy_addr is 0x%08x\n",td->mmap_user[index].phy_addr);
	return (td->mmap_user[index].phy_addr);
}

mpm_transport_trans_h mpm_transport_hyplnk_get_initiate64(mpm_transport_cfg_t *sp, uint64_t from_addr, uint64_t to_addr, uint32_t length, bool is_blocking, mpm_transport_read_t *cfg)
{
	mpm_transport_trans_t *tp = NULL;
	int i;
	void *mmap_addr;
	uint32_t logical_addr;
	
	debug_printf("Base = 0x%08x\n",(sp->mpax_base));
	debug_printf("Size = 0x%08x\n",(sp->mpax_size));

	mpm_transport_mmap_t mcfg = {
		.mmap_prot	= (PROT_READ|PROT_WRITE),
		.mmap_flags	= MAP_SHARED,
	};
	for (i = 0; i < MPM_MAX_DMA_TRANS; i++)
	{
		if (sp->ta[i].is_active == 0)
			break;
	}
	if (i == MPM_MAX_DMA_TRANS)
	{
		printf("MAX DMA Transfers! Call mpm_transport_transfer_check to wait/ccomplete pending transfers\n");
		tp->key = -1;
		return tp;
	}
	tp = (mpm_transport_trans_t *) &(sp->ta[i]);
	tp->key = (i+1) * (i+1) * 2;
	tp->is_active = 1;

	mmap_addr = (void *) mpm_transport_hyplnk_phys_map64 (sp, from_addr, length, &mcfg);
	if (!mmap_addr) {
		printf("can't map for address to_addr 0x%llx\n", to_addr);
		tp->key = -1;
		return tp;
	}
	debug_printf("get_initiate64: Got mmap_addr = 0x%08x\n",(uint32_t)mmap_addr);
	tp->trans_addr = (uint32_t) mmap_addr;
	
	if (!(sp->hMpax)) {
		sp->hMpax = mpm_transport_keystone_mmap_open(sp->mpax_base, sp->mpax_size, sp->mpax_index, sp->mpax_count, 0, 0, 8);
	}
	logical_addr = mpm_transport_keystone_mmap(sp->hMpax, to_addr, length);
	if (logical_addr < 0) {
		printf("can't map for address from_addr 0x%llx\n", to_addr);
		tp->key = -1;
		return tp;
	}
	debug_printf("get_initiate64: Got logical = 0x%08x\n",logical_addr);
	tp->local_logical_addr = logical_addr;

	debug_printf("mpm_transport_get_initiate64: getting address 0x%llx, to 0x%llx\n", from_addr, to_addr);
	mpm_transport_edma3_transfer(sp->hEdma[sp->edma3_inst], (uint32_t) mmap_addr, logical_addr, length, is_blocking, tp);
	debug_printf("hEdma = 0x%08x, Channel id: 0x%08x\n", (uint32_t) sp->hEdma[sp->edma3_inst], tp->chan_id);
	if (is_blocking == true) {
		mpm_transport_hyplnk_phys_unmap(sp, mmap_addr, 0);
		mpm_transport_keystone_unmap(sp->hMpax,logical_addr,sp->mpax_size);
	}
	return tp;
}

mpm_transport_trans_h mpm_transport_hyplnk_put_initiate64(mpm_transport_cfg_t *sp, uint64_t to_addr, uint64_t from_addr, uint32_t length, bool is_blocking, mpm_transport_write_t *cfg)
{
	mpm_transport_trans_t *tp = NULL;
	int i;
	void *mmap_addr;
	uint32_t logical_addr;
	
	debug_printf("Base = 0x%08x\n",(sp->mpax_base));
	debug_printf("Size = 0x%08x\n",(sp->mpax_size));

	mpm_transport_mmap_t mcfg = {
		.mmap_prot	= (PROT_READ|PROT_WRITE),
		.mmap_flags	= MAP_SHARED,
	};
	for (i = 0; i < MPM_MAX_DMA_TRANS; i++)
	{
		if (sp->ta[i].is_active == 0)
			break;
	}
	if (i == MPM_MAX_DMA_TRANS)
	{
		printf("MAX DMA Transfers! Call mpm_transport_transfer_check to wait/ccomplete pending transfers\n");
		tp->key = -1;
		return tp;
	}
	tp = (mpm_transport_trans_t *) &(sp->ta[i]);
	tp->key = (i+1) * (i+1) * 2;
	tp->is_active = 1;

	mmap_addr = (void *) mpm_transport_hyplnk_phys_map64 (sp, to_addr, length, &mcfg);
	if (!mmap_addr) {
		printf("can't map for address to_addr 0x%llx\n", to_addr);
		tp->key = -1;
		return tp;
	}
	debug_printf("put_initiate64: Got mmap_addr = 0x%08x\n",(uint32_t)mmap_addr);
	tp->trans_addr = (uint32_t) mmap_addr;

	if (!(sp->hMpax)) {
		sp->hMpax = mpm_transport_keystone_mmap_open(sp->mpax_base, sp->mpax_size, sp->mpax_index, sp->mpax_count, 0, 0, 8);
	}
	logical_addr = mpm_transport_keystone_mmap(sp->hMpax, from_addr, length);
	if (logical_addr < 0) {
		printf("can't map for address from_addr 0x%llx\n", from_addr);
		tp->key = -1;
		return tp;
	}
	debug_printf("put_initiate64: Got logical = 0x%08x\n",logical_addr);
	tp->local_logical_addr = logical_addr;

	debug_printf("mpm_transport_put_initiate64: writing to address 0x%llx, from 0x%llx\n", to_addr, from_addr);
	mpm_transport_edma3_transfer(sp->hEdma[sp->edma3_inst], logical_addr, (uint32_t) mmap_addr, length, is_blocking, tp);
	debug_printf("hEdma = 0x%08x, Channel id: 0x%08x\n", (uint32_t) sp->hEdma[sp->edma3_inst], tp->chan_id);
	if (is_blocking == true) {
		mpm_transport_hyplnk_phys_unmap(sp, mmap_addr, 0);
		mpm_transport_keystone_unmap(sp->hMpax,logical_addr,sp->mpax_size);
	}
	return tp;
}
